#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Wind- and turbulence-related claculations and conversions
"""
import logging

import numpy as np

from ._utils import _check, _only
from .constants import gn, kappa

logger = logging.getLogger(__name__)

DISPLACEMENT_FACTOR = 6.5


class LogWind(object):
    '''
    The class LogWind represents a logarithmic wind profile that
    my be described and requested in any possible compination of
    parameters.

    :param ust: friction velocity in m/s (float).
    :param z0:  surface rougness length in m (float).
    :param d:   displacement height in m (float). Defaults to 0 m.
    :param u:   wind speed at height `z` in m/s  (float).
    :param z:   height above ground, where `u` is measured, in m (float).
    :param u2:   wind speed at height `z2` in m/s  (float).
    :param z2:   height above ground, where `u2` is measured, in m (float).

    Possible parameter combinations for initialization are:
      - `ust` and `z0`, optionally `d`
      - `u`, `z` and `z0`, optionally `d`
      - `u`, `z` and `ust`, optionally `d`
      - `u`, `z`, `u2`, and `z2`, optionally `d`

    :ivar ust: friction velocity in m/s
    :ivar z0: surface rougness length in m
    :ivar ~.d: displacement height in m


    '''

    def __init__(self, ust=None, z0=None, d=None, u=None, z=None,
                 u2=None, z2=None):
        loc = locals()
        par = {x: loc[x] for x in ['ust', 'z0', 'd', 'u', 'z', 'u2', 'z2']}

        if _only(par, ['ust', 'z0'], ['d']):
            ust = _check('ust', ust, 'float', ge=0.)
            self.ust = ust
            z0 = _check('z0',  z0, 'float',  gt=0.)
            self.z0 = z0
            d = _check('d',    d, 'float',  ge=0., none=True)
            if d is None:
                self.d = DISPLACEMENT_FACTOR * self.z0
            else:
                self.d = d  # : displacement height in m

        elif _only(par, ['u', 'z', 'z0'], ['d']):
            z0 = _check('z0',  z0, 'float',  gt=0.)
            self.z0 = z0
            d = _check('d',   d, 'float',  ge=0., none=True)
            if d is None:
                self.d = DISPLACEMENT_FACTOR * self.z0
            else:
                self.d = d
            u = _check('u',   u, 'float',   ge=0.)
            z = _check('z',   z, 'float',   ge=self.d)
            self.ust = (kappa * u) / np.log((z-self.d)/self.z0)

        elif _only(par, ['u', 'z', 'ust'], ['d']):
            ust = _check('ust', ust, 'float', ge=0.)
            self.ust = ust
            u = _check('u',     u, 'float', gt=0.)
            d = _check('d',     d, 'float', ge=0., none=True)
            if d is None:
                z = _check('z',   z, 'float', gt=0.)
                self.z0 = z / (DISPLACEMENT_FACTOR + np.exp(kappa * u / ust))
                self.d = DISPLACEMENT_FACTOR * z0
            else:
                z = _check('z',   z, 'float', gt=self.d)
                self.z0 = (z-d) / np.exp(kappa * u / ust)
                self.d = d

        elif _only(par, ['u', 'z', 'u2', 'z2'], ['d']):
            d = _check('d',     d, 'float', ge=0., none=True)
            if d is None:
                u = _check('u',   u, 'float', ge=0.)
                z = _check('z',   z, 'float', ge=0.)
                u2 = _check('u2', u2, 'float', gt=u)
                z2 = _check('z2', z2, 'float', gt=z)
                z0l = np.min([0.5 * z, 0.01])
                if z <= DISPLACEMENT_FACTOR * z0l:
                    raise ValueError('z too low')
                for i in range(100):
                    logger.debug(format((z0l, DISPLACEMENT_FACTOR*z0l, z)))
                    zz = np.exp((u*np.log(z2-DISPLACEMENT_FACTOR*z0l) -
                                u2*np.log(z-DISPLACEMENT_FACTOR*z0l)) /
                                (u - u2))
                    z0 = z0l * (zz/z0l)**0.25
                    if abs(np.log(z0/z0l)) < 0.001:
                        break
                    z0l = z0
                else:
                    raise ValueError('did not converge')
                self.z0 = z0
                self.d = DISPLACEMENT_FACTOR*z0
                self.ust = u / np.log((z-self.d) / self.z0)
            else:
                self.d = d
                u = _check('u',   u,  'float', ge=0.)
                z = _check('z',   z,  'float', ge=self.d)
                u2 = _check('u2',  u2, 'float', gt=u)
                z2 = _check('z2',  z2, 'float', gt=z)
                if z <= d + d / DISPLACEMENT_FACTOR:
                    raise ValueError('z below displacement')
                self.z0 = np.exp(
                    (u*np.log(z2-self.d) - u2*np.log(z-self.d)) / (u - u2))
                self.ust = u / np.log((z-self.d) / self.z0)

    def u(self, z):
        '''
        returns wind speed at certain level

        :param z: height above ground in m (float)
        :return: wind speed in m/s
        :rtype: float
        '''
        z = _check('z', z, 'float', ge=0.)
        u = (self.ust / kappa) * np.log((z-self.d)/self.z0)
        return u

    def gradu(self, z):
        '''
        returns wind speed vertical gradient at certain level

        :param z: height above ground in m (float)
        :return: wind speed gradient  in 1/s
        :rtype: float
        '''
        z = _check('z', z, 'float', ge=0.)
        u = self.ust / (kappa * (z - self.d))
        return u

# -------------------------------------------------------------------


class DiabaticWind(object):
    '''
    The class LogWind represents a logarithmic wind profile that
    my be described and requested in any possible compination of
    parameters.

    :param ust: friction velocity in m/s (float).
    :param z0:  surface rougness length in m (float).
    :param d:   displacement height in m (float). Defaults to 0 m.
    :param u:   wind speed at height `z` in m/s  (float).
    :param z:   height above ground, where `u` is measured, in m (float).
    :param zoL: Monin-Boukhov stability parameter z/L, unitless (float).

    Possible parameter combinations for initialization are:
      - `ust` and `z0`, 'z', 'L', optionally `d`
      - `u`, `z` and `z0`, optionally `d`
      - `u`, `z` and `ust`, optionally `d`
      - `u`, `z`, `u2`, and `z2`, optionally `d`

    :ivar ust: friction velocity in m/s
    :ivar z0: surface rougness length in m
    :ivar ~.d: displacement height in m
    :ivar LOb: Obukhov length in m


    '''

    def __init__(self, ust=None, z0=None, d=0., u=None, z=None,
                 LOb=None, zoL=None):
        loc = locals()
        par = {x: loc[x] for x in ['ust', 'z0', 'd', 'u', 'z', 'LOb', 'zoL']}

        if _only(par, ['ust', 'z0', 'LOb'], ['d']):
            ust = _check('ust', ust, 'float', ge=0.)
            self.ust = ust
            z0 = _check('z0',  z0, 'float',  gt=0.)
            self.z0 = z0
            LOb = _check('LOb', LOb, 'float')
            self.LOb = LOb
            d = _check('d',    d, 'float',  ge=0., none=True)
            if d is None:
                self.d = DISPLACEMENT_FACTOR * self.z0
            else:
                self.d = d  # : displacement height in m

        elif _only(par, ['u', 'z', 'z0', 'LOb'], ['d']):
            z0 = _check('z0',  z0, 'float',  gt=0.)
            self.z0 = z0
            d = _check('d',   d, 'float',  ge=0., none=True)
            if d is None:
                self.d = DISPLACEMENT_FACTOR * self.z0
            else:
                self.d = d
            u = _check('u',   u, 'float',   ge=0.)
            z = _check('z',   z, 'float',   ge=self.d)
            self.ust = (kappa * u) / np.log((z-self.d)/self.z0)
            LOb = _check('LOb', LOb, 'float')
            self.LOb = LOb
            zoL = z / LOb
            self.ust = ((kappa * u) /
                        (np.log((z-self.d)/self.z0) - psi_m(zoL)))

        elif _only(par, ['u', 'z', 'z0', 'zoL'], ['d']):
            z0 = _check('z0',  z0, 'float',  gt=0.)
            self.z0 = z0
            d = _check('d',   d, 'float',  ge=0., none=True)
            if d is None:
                self.d = DISPLACEMENT_FACTOR * self.z0
            else:
                self.d = d
            u = _check('u',   u, 'float',   ge=0.)
            z = _check('z',   z, 'float',   ge=self.d)
            zoL = _check('zoL', zoL, 'float')
            self.LOb = z/zoL
            self.ust = ((kappa * u) /
                        (np.log((z-self.d)/self.z0) - psi_m(zoL)))

#    elif _only(par,['u','z','ust',],['d']):
#      ust = _check('ust', ust, 'float', ge=0.)
#      self.ust = ust
#      u   = _check('u',     u, 'float', gt=0.)
#      d   = _check('d',     d, 'float', ge=0., none=True)
#      if d is None:
#        z   = _check('z',   z, 'float', gt=0.)
#        self.z0 = z  / ( DISPLACEMENT_FACTOR + np.exp( kappa * u / ust) )
#        self.d = DISPLACEMENT_FACTOR * z0
#      else:
#        z   = _check('z',   z, 'float', gt=self.d)
#        self.z0 = (z-d) / np.exp( kappa * u / ust)
#        self.d = d

    def u(self, z):
        '''
        returns wind speed at certain level

        :param z: height above ground in m (float)
        :return: wind speed in m/s
        :rtype: float
        '''
        z = _check('z', z, 'float', ge=0.)
        zoL = z / self.LOb
        u = (self.ust / kappa) * (np.log((z-self.d) /
                                         self.z0) - psi_m(zoL))
        return u

    def gradu(self, z):
        '''
        returns wind speed vertical gradient at certain level

        :param z: height above ground in m (float)
        :return: wind speed gradient  in 1/s
        :rtype: float
        '''
        z = _check('z', z, 'float', ge=0.)
        zoL = z / self.LOb
        gu = psi_m(zoL) * self.ust / (kappa * (z - self.d))
        return gu

# -------------------------------------------------------------------


def phi_m(zoL):
    '''
    Universal functions for momentum according to
    [Bus1971]_, as recalculated by [Hog1988]_.

    :param zoL: Monin-Boukhov stability parameter z/L, unitless (float).
    :return: univesal fuction value (unitless)
    :rtype: float
    '''
    zoL = _check('zoL', zoL, 'float')
    phi_1s = 6.0
    phi_1u = 19.3
    if zoL > 0.:
        phi = 1. + phi_1s * zoL
    elif zoL < 0.:
        phi = (1. - phi_1u * zoL) ^ (-1./4.)
    else:  # if zoL == 0.
        phi = 1.
    return phi

# -------------------------------------------------------------------


def phi_H(zoL):
    '''
    Universal functions for turbulent fluxes of scalars according to
    [Bus1971]_, using the numerical "modified Kansas expression"
    valuesas recalculated by [Hog1988]_.

    :param zoL: Monin-Boukhov stability parameter z/L, unitless (float).
    :return: univesal fuction value (unitless)
    :rtype: float
    '''
    zoL = _check('zoL', zoL, 'float')
    phi_1s = 7.8
    phi_1u = 11.6
    phi_0 = 0.95
    if zoL > 0.:
        phi = phi_0 + phi_1s * zoL
    elif zoL < 0.:
        phi = phi_0 * (1. - phi_1u * zoL) ^ (-1./2.)
    else:  # if zoL == 0.
        phi = phi_0
    return phi

# -------------------------------------------------------------------


def psi_m(zoL):
    '''
    Integrated universal functions for momentum according to
    [Pau1970]_ (unstable) and [HoB1988]_ (stable),
    using the numerical "modified Kansas expression"
    values recalculated by [Hog1988]_
    :param zoL: Monin-Boukhov stability parameter z/L, unitless (float).
    :return: universal fuction value (unitless)
    :rtype: float
    '''
    zoL = _check('zoL', zoL, 'float')
    phi_1s = 6.0
    phi_1u = 19.3
    if zoL > 0.:
        psi = - phi_1s * zoL
    else:
        x = (1 - phi_1u * zoL) ** 0.25
        psi = (2. * np.log((1. + x) / 2.)
               + np.log((1. + x ** 2) / 2.)
               - 2. * np.arctan(x)
               + np.pi / 2.)
    return psi

# -------------------------------------------------------------------


def psi_H(zoL):
    '''
    Integrated universal functions for momentum according to
    [Pau1970]_ (unstable) and [HoB1988]_ (stable),
    using the numericcal "modified Kansas expression"
    values by [Hog1988]_
    :param zoL: Monin-Boukhov stability parameter z/L, unitless (float).
    :return: univesal fuction value (unitless)
    :rtype: float
    '''
    zoL = _check('zoL', zoL, 'float')
    phi_2s = 7.8
    phi_2u = 11.6
    phi_20 = 0.95
    if zoL > 0.:
        psi = - phi_2s * zoL
    else:
        x = (1 - phi_2u * zoL) ** 0.25
        psi = phi_20 * (2. * np.log((1. + x ** 2) / 2.))
    return psi


# -------------------------------------------------------------------
def charnock(ust):
    '''
    Charnock's relation: water surface roughness
    :param ust: friction velocity in m/s (float).
    :return: water surface roughness length in m
    :rtype: float
    '''
    ust = _check('ust', ust, 'float', ge=0.)
    alpha_c = 0.015  # Charnock parameter
    z0 = alpha_c * (ust**2) / gn
    return z0

# -------------------------------------------------------------------


def transfer(wind, z0, d=0.):
    '''
    returns wind profile for a different place with the same
    synoptic conditions but different roughness (and displacement) length(s)
    :param wind: logarithmic wind profile at **old** location (LogWind).
    :param z0: surface roughness length at **new** location in m (float).
    :param z0: (optional) displacement height at **new** location in m (float).
    :return: logarithmic wind profile at **new** location
    :rtype: LogWind
    '''
    if not isinstance(wind, LogWind):
        raise TypeError('agrmument wind must be of type LogWind')
    Refheight = 250.  # m
    uref = wind.u(Refheight)
    new = LogWind(u=uref, z=Refheight, z0=z0, d=d)
    return new

# -------------------------------------------------------------------


def vectormean(ff=None, dd=None, pairs=None):
    '''
    returns vector mean of wind vectors given by means of speed an direction
    or lists of speed and direction

    :param ff: wind speed in any unit,
      requires `dd` (array-like, same lenght as `dd`)
    :param dd: wind direction in degrees,
      requires `ff` (array-like, same lenght as `ff`)
    :param pairs: pairs of wind speed (any unit) and direction (in degrees)
      (list of tuples)
    :return: mean wind speed and wind direction in degrees
    :rtype: float, float
    '''
    if pairs is not None:
        ff = [x[0] for x in pairs]
        dd = [x[1] for x in pairs]
    elif ff is not None and dd is not None:
        if len(ff) != len(dd):
            raise ValueError('ff and dd mut be of same length')
    else:
        raise ValueError('einter pairs or ff and dd must be given')

    u = [-f*np.sin(np.deg2rad(d)) for f, d in zip(ff, dd)]
    v = [-f*np.cos(np.deg2rad(d)) for f, d in zip(ff, dd)]
    umean = np.nanmean(u)
    vmean = np.nanmean(v)

    fm = np.sqrt(umean**2 + vmean**2)
    dm = np.rad2deg(np.arctan2(-umean, -vmean))

    return fm, dm

# -------------------------------------------------------------------


def uv2dir(u, v, pos=True):
    '''
    converts (horizontal) wind vector components to speed and direction

    :param u: eastward wind component (float or array-like)
    :param v: northward wind component (float or array-like)
    :param pos: (optional) return positive values:
        If ``True`` returned wind directions are in the range 0..360,
        if ``False`` values are in the range -180..180.
        Defaults to ``True``.
    :return: wind speed, same unit as input, and wind direction in degrees
    :rtype: float, float or np.array, np.array
    '''
    # test if shapes are identical
    if np.shape(u) != np.shape(v):
        raise ValueError('u and v must be of the same shape')
    # convert scalars into array
    if np.isscalar(u) and np.isscalar(v):
        u = np.array([u])
        v = np.array([v])
    else:
        u = np.array(u)
        v = np.array(v)
    # calculate
    speed = np.sqrt(u**2 + v**2)
    wdir = np.rad2deg(np.arctan2(-u, -v))
    # make values (range -180..180) positive (0..360)
    if pos is True:
        wdir = wdir % 360
    # convert single values back to scalars
    if speed.shape == (1,):
        speed = speed[0]
        wdir = wdir[0]
    return speed, wdir

# -------------------------------------------------------------------


def dir2uv(ff, dd):
    '''
    converts (horizontal) wind vector from speed and direction
    to vector components

    :param ff: wind speed in any unit (float or array-like)
    :param dd: wind direction in degrees (float or array-like)
    :return: eastward wind component and northward wind component
        in the same unit `speed` is given in
    :rtype: float, float or np.array, np.array
    '''
    # test if shapes are identical
    if np.shape(ff) != np.shape(dd):
        raise ValueError('ff and dd must be of the same shape')
    # convert scalars into array
    if np.isscalar(ff) and np.isscalar(dd):
        ff = np.array([ff])
        dd = np.array([dd])
    else:
        ff = np.array(ff)
        dd = np.array(dd)
    # check values:
    if any(ff < 0.):
        raise ValueError('ff must be positive')
    # calculate
    u = - ff * np.sin(np.deg2rad(dd))
    v = - ff * np.cos(np.deg2rad(dd))
    # convert single values back to scalars
    if u.shape == (1,):
        u = u[0]
        v = v[0]
    return u, v



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 29 02:13:26 2019

@author: clemens
"""

import unittest

import meteolib as m


class Test_gas_p(unittest.TestCase):
    '''
    def gas_p(rho, T, q=None, Kelvin=None, hPa=False, gas='air')
    '''

    def test_gas_p_dry(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, gas='dry'), 98009.6, delta=1.)

    def test_gas_p_moist(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, q=0, gas='moist'), 98009.6, delta=1.)
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, q=2.E-3, gas='moist'), 98129.2, delta=1.)

    def test_gas_p_water(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, gas='water'), 157576.8, delta=1.)

    def test_gas_p_co2(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, gas='co2'), 64504.4, delta=1.)

    def test_gas_p_Celsius(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 0., Kelvin=False, gas='dry'), 98009.6, delta=1.)

    def test_gas_p_hPa(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, hPa=True, gas='dry'), 980.096, delta=0.01)

    def test_gas_p_gkg(self):
        self.assertAlmostEqual(m.thermodyn.gas_p(
            1.25, 273.15, q=2., gkg=True, gas='moist'), 98129.2, delta=1.)


class Test_gas_rho(unittest.TestCase):
    '''
    def gas_rho(p, T, q=None, Kelvin=None, hPa=False, gas='air')
    '''

    def test_gas_rho_dry(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            98009.6, 273.15, gas='dry'), 1.25, delta=0.01)

    def test_gas_rho_moist(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            98009.6, 273.15, q=0, gas='moist'), 1.25, delta=0.01)
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            98129.2, 273.15, q=2.E-3, gas='moist'), 1.25, delta=0.01)

    def test_gas_rho_water(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            157576.8, 273.15, gas='water'), 1.25, delta=0.01)

    def test_gas_rho_co2(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            64504.4, 273.15, gas='co2'), 1.25, delta=0.01)

    def test_gas_rho_Celsius(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            98009.6, 0., Kelvin=False, gas='dry'), 1.25, delta=0.01)

    def test_gas_rho_hPa(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            980.096, 273.15, hPa=True, gas='dry'), 1.25, delta=0.01)

    def test_gas_rho_gkg(self):
        self.assertAlmostEqual(m.thermodyn.gas_rho(
            98129.2, 273.15, q=2., gkg=True, gas='moist'), 1.25, delta=0.01)


class Test_cpt(unittest.TestCase):
    '''
    def cpt(t, Kelvin=None):
    '''

    def test_cpt_auto(self):
        self.assertAlmostEqual(m.thermodyn.cpt(0.), m.constants.cp, delta=1.)


class Test_cpvt(unittest.TestCase):
    '''
    def cpvt(t, Kelvin=None):
    '''

    def test_cpvt_auto(self):
        self.assertAlmostEqual(m.thermodyn.cpvt(
            0.), m.constants.cpv, delta=10.)


class Test_cwt(unittest.TestCase):
    '''
    def cwt(t, Kelvin=None):
    '''

    def test_cwt_auto(self):
        self.assertAlmostEqual(m.thermodyn.cwt(
            0.), m.constants.cw, delta=0.01*m.constants.cw)


class Test_cit(unittest.TestCase):
    '''
    def cit(t, Kelvin=None):
    '''

    def test_cit_auto(self):
        self.assertAlmostEqual(m.thermodyn.cit(
            0.), m.constants.ci, delta=0.015*m.constants.ci)


class Test_Lvt(unittest.TestCase):
    '''
    def Lvt(t, Kelvin=None):
    '''
    # Smithonian tabels (1951), Table 92 (in cal/gK; J = 4.1839 cal)
    t92 = {
        0: 597.31,
        10: 591.7,
        20: 586.0,
        30: 580.4,
        40: 574.7,
        50: 569.0,
    }
    j92 = 4.1839E3

    def test_Lvt_auto(self):
        for t, c in self.t92.items():
            with self.subTest(t=t):
                self.assertAlmostEqual(m.thermodyn.Lvt(
                    t), c*self.j92, delta=0.005*c*self.j92)

    def test_Lvt_sources(self):
        for s in ['andreas', 'henderson-sellers',
                  'ecpack', 'dake', 'hyland-wexler']:
            for t, c in self.t92.items():
                with self.subTest(s=s, t=t):
                    self.assertAlmostEqual(
                        m.thermodyn.Lvt(t, source=s),
                        c*self.j92, delta=0.0075*c*self.j92)


class Test_Lst(unittest.TestCase):
    '''
    def Lst(t, Kelvin=None):
    '''

    def test_Lst_sources(self):
        for s in ['andreas', 'hyland-wexler']:
            t = 0
            with self.subTest(s=s, t=t):
                self.assertAlmostEqual(m.thermodyn.Lst(
                    t, source=s), m.constants.Ls, delta=0.005*m.constants.Ls)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Humidity calculations and conversions
'''
import numpy as np

from ._utils import _check, _only
from .constants import Lv, Ttriple, cp, pzero
from .pressure import _to_Pa, _to_hPa
from .temperature import Tpot, _to_C, _to_K
from .thermodyn import gas_p, gas_rho

_esat = 'magnus'

# ---------------------------------------------------------------------


def _f(p):
    '''
    Moist air correction for water vapor saturation pressure
    :param p: air pressure in hPa (float)
    :return: correction factor. If `p` is ``None``, the value 1. is returned.

    This form of the equation was adopted by Forty-second session of the
    Executive Council (EC-XLII) in 1990 [WMO8]_:
    .. math::
      f (p) = 1.001 6 + 3.15 · 10 –6 p – 0.074 p –1

    Usage::

      >>> _f(1013.)
      1.001

    '''
    if p is None:
        f = 1.
    else:
        f = 1.0016 + 3.15E-6 * p - 0.074 / p
    return (f)

# ---------------------------------------------------------------------


def magnus_w(t, Kelvin=None, p=None, hPa=False):
    '''
    Magnus Formula, returns the saturation pressure for water vapor
    over an even liquid water surface in hPa

    :param t: Temperature
    :param Kelvin: (optional)
      if ``True``, unit of `t` is assumed to be Kelvin.
      If ``False``, unit of `t` is assumed to be Celsius.
      If missing of ``None``, unit of `t` is autodetected. Defaults to
      ``None``.
    :param p: (optional) air pressure (float). If `p` is ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``False``, `p` must be supplied and `magnus_w is returned in Pa .
      If ``True``, `p` must be supplied and `magnus_w is returned in hPa.
      If missing or ``None`` the unit of `p` is autodetected
      and `magnus_w is returned in Pa. Defaults to ``None``.
    :return: saturation pressure in hPa. If `t` is not a valid number,
      ``np.nan`` is returned.
    :rtype: float
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, ``None``,
      or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, ``None``,
      or missing

    This form of the equation was adopted by Forty-second session of the
    Executive Council (EC-XLII) in 1990 [WMO8]_:

    .. math::
      e = f(p) * 6.112 * np.exp((17.62 * t) / (243.12 + t))
      :label: magnus

    Results are valid in the range water -45°C to +60°C for pure phase water.

    '''
    t = _to_C(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        e = 6.112 * np.exp((17.62 * t) / (243.12 + t))  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def magnus_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Magnus Formula, returns the saturation pressure for water vapor
    over an even liquid water surface in hPa

    :param t: Temperature
    :param Kelvin: (optional)
      if ``True``, unit of  `t` is assumed to be Kelvin.
      If ``False``, unit of  `t` is assumed to be Celsius.
      If missing or ``None``, unit of  `t` is autodetected. Defaults to
      ``None``.
    :param p: (optional) air pressure (float). If `p` is ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``, `p` must be supplied and `magnus_i` is returned in hPa.
      If ``False``, `p` must be supplied and `magnus_i` is returned in Pa .
      If missing or ``None`` the unit of `p` is autodetected
      and `magnus_i` is returned in Pa. Defaults to ``None``.
    :return: saturation pressure in hPa. If `t` is not a valid number,
      ``np.nan`` is returned.
    :rtype: float
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, ``None``,
      or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, ``None``,
      or missing

    This form of the equation was adopted by Forty-second session of the
    Executive Council (EC-XLII) in 1990 [WMO8]_:

    .. math::
      e = 6.112 * np.exp((22.46 * t) / (272.62 + t))
      :label: magnice

    Results are valid in the range water -65°C to 0°C for pure phase water.

    '''
    t = _to_C(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        e = 6.112 * np.exp((22.46 * t) / (272.62 + t))  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def tetens_w(t, Kelvin=None, p=None, hPa=False):
    '''
    Tetens formula, identical to :py:meth:`meltib.humidity.magnus_w`, except
    Coefficients provided by [Tet1930]_, cited from [Mon2008]_:

    .. math::


      e = 6.1078 * np.exp((17.27 * t) / (237.3 + t))

    Results are within 1 Pa of exact values in the range -0°C to +35°C
    [Mon2008]_.

    '''
    t = _to_C(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        e = 6.1078 * np.exp((17.27 * t) / (237.3 + t))  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def tetens_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Tetens formula, identical to :py:meth:`meltib.humidity.magnus_i`, except
    Coefficients provided by [Tet1930]_, cited from [Mon2008]_:

    .. math::


      e = 6.1078 * np.exp((21.875 * t) / (265.5 + t))

    Results are within 1 Pa of exact values in the range -20°C to +0°C
    [Mon2008]_.

    '''
    t = _to_C(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        e = 6.1078 * np.exp((21.875 * t) / (265.5 + t))  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def goff_gratch_w(t, Kelvin=None, p=None, hPa=False):
    '''
    Goff-Gratch formula, identical to :py:meth:`meltib.humidity.magnus_w`,
    except formula provided by [Gof1957]_, cited from [WMO49]_:

    .. math::

      log10(ew) = & 10.79574 * (1 – Ttriple/T) \\\\
                  &– 5.02800 * np.log10 (T/Ttriple) \\\\
                  &+ 1.50475E–4 * (1 – 10**(–8.2969*(T/Ttriple–1)) ) \\\\
                  &+ 0.42873E–3 * ( 10**(4.76955*(1-Ttriple/T)) – 1) \\\\
                  &+ 0.78614

    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(T):
        e = np.nan
    else:
        loge = (10.79574 * (1 - Ttriple/T)
                - 5.02800 * np.log10(T/Ttriple)
                + 1.50475E-4 * (1 - 10**(-8.2969*(T/Ttriple-1)))
                + 0.42873E-3 * (10**(4.76955*(1-Ttriple/T)) - 1)
                + 0.78614
                )
        e = 10**loge
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def goff_gratch_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Goff-Gratch formula, identical to :py:meth:`meltib.humidity.magnus_i`,
    except formula provided by [Gof1957]_, cited from [WMO49]_:

    .. math::

      log10(ei) = & -9.09718 * (Ttriple/T - 1) \\\\
                  &- 3.56654 * np.log10(Ttriple/ T) \\\\
                  &+ 0.876793 * (1 - T/Ttriple) \\\\
                  &+ np.log10(6.1071) \\\\
                  )

    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        loge = (-9.09718 * (Ttriple/T - 1)
                - 3.56654 * np.log10(Ttriple / T)
                + 0.876793 * (1 - T/Ttriple)
                + np.log10(6.1071)
                )
        e = 10**loge
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def hyland_wexler_w(t, Kelvin=None, p=None, hPa=False):
    '''
    Formula by Hyland and Wexler,
    identical to :py:meth:`meltib.humidity.magnus_w`,
    except formula provided by [HyW1983]_:

    .. math::

      ln(ew)    =  & - 0.58002206E4 / T \\\\
                   & + 0.13914993E1 \\\\
                   & - 0.48640239E-1 * T \\\\
                   & + 0.41764768E-4 * T^2 \\\\
                   & - 0.14452093E-7 * T^3 \\\\
                   & + 0.65459673E1 * ln(T)
    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(T):
        e = np.nan
    else:
        lne = (- 0.58002206E4 / T
               + 0.13914993E1
               - 0.48640239E-1 * T
               + 0.41764768E-4 * T**2
               - 0.14452093E-7 * T**3
               + 0.65459673E1 * np.log(T)
               )
        e = np.exp(lne)  # Pa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e / 100.  # hPa
    else:
        return e  # Pa

# ---------------------------------------------------------------------


def hyland_wexler_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Formula by Hyland and Wexler,
    identical to :py:meth:`meltib.humidity.magnus_i`,
    except formula provided by [HyW1983]_:

    .. math::

      ln(ei)    = &- 0.56745359E4 / T \\\\
                  &+ 0.63925247E1 \\\\
                  &- 0.96778430E-2 * T \\\\
                  &+ 0.62215701E-6 * T^2 \\\\
                  &+ 0.20747825E-8 * T^3 \\\\
                  &- 0.94840240E-12 * T^4 \\\\
                  &+ 0.41635019E1 * ln(T) \\\\

    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        lne = (- 0.56745359E4 / T
               + 0.63925247E1
               - 0.96778430E-2 * T
               + 0.62215701E-6 * T**2
               + 0.20747825E-8 * T**3
               - 0.94840240E-12 * T**4
               + 0.41635019E1 * np.log(T)
               )
        e = np.exp(lne)  # Pa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e / 100.  # hPa
    else:
        return e  # Pa

# ---------------------------------------------------------------------


def sonntag_w(t, Kelvin=None, p=None, hPa=False):
    '''
    Formula by Sonntag, identical to :py:meth:`meltib.humidity.magnus_w`,
    except formula provided by [Son1990]_
    (and [Son1994]_,but with fewer digits):

    .. math::

      ln(ew)    =  & - 6096.9385 / T \\\\
                    + 16.635794 \\\\
                    - 2.711193E-2 * T \\\\
                    + 1.673952E-5 * T**2 \\\\
                    + 2.433502 * ln(T) \\\\
    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(T):
        e = np.nan
    else:
        lne = (- 6096.9385 / T
               + 16.635794
               - 2.711193E-2 * T
               + 1.673952E-5 * T**2
               + 2.433502 * np.log(T)
               )
        e = np.exp(lne)  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def sonntag_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Formula by Sonntag, identical to :py:meth:`meltib.humidity.magnus_i`,
    except formula provided by [Son1990]_
    (and [Son1994]_, but with fewer digits):

    .. math::

      ln(ei)    = &- 6024.5282 / T \\\\
                  &+ 24.721994 \\\\
                  &+ 1.0613868E-2 * T \\\\
                  &- 1.3198825E-5 * T^2 \\\\
                  &- 0.49382577 * ln(T) \\\\

    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(t):
        e = np.nan
    else:
        lne = (- 6024.5282 / T
               + 24.721994
               + 1.0613868E-2 * T
               - 1.3198825E-5 * T**2
               - 0.49382577 * np.log(T)
               )
        e = np.exp(lne)  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa


# ---------------------------------------------------------------------
def iapws_w(t, Kelvin=None, p=None, hPa=False):
    '''
    International Association for the Properties of Water and Steam (IAPWS)
    1995 Formula for water vapor,
    published by Wagner and Pruss [WaP2002]_ eqn. 2,5a:

    .. math::

      Log (ew/22.064E6) = 647.096/T * ((-7.85951783 v
          + 1.84408259 * v**1.5
          - 11.7866497 * v**3
          + 22.6807411 * v**3.5
          - 15.9618719 * v**4
          + 1.80122502 * v**7.5))


    with T in [K] and ew in [Pa] and v = 1 - T/647.096
    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(T):
        e = np.nan
    else:
        v = 1. - T / 647.096
        lne = 647.096/T * (
            - 7.85951783 * v
            + 1.84408259 * v**1.5
            - 11.7866497 * v**3
            + 22.6807411 * v**3.5
            - 15.9618719 * v**4
            + 1.80122502 * v**7.5
        )
        e = np.exp(lne) * 22.064E4  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa

# ---------------------------------------------------------------------


def iapws_i(t, Kelvin=None, p=None, hPa=False):
    '''
    Sublimation pressure equation by Wagner, as published by Wagner and Pruss
    [WaP2002]_ eqn. 2.21, matching the
    International Association for the Properties of Water and Steam (IAPWS)
    1995 Formula

    .. math::


      Log (ew/22.064E6) = 647.096/T * ((-7.85951783 v
          + 1.84408259 * v**1.5
          - 11.7866497 * v**3
          + 22.6807411 * v**3.5
          - 15.9618719 * v**4
          + 1.80122502 * v**7.5))

      with T in [K] and ew in [Pa] and v = 1 - T/647.096
    '''
    T = _to_K(t, Kelvin)  # K
    if np.isnan(T):
        e = np.nan
    else:
        th = T / 273.16  # 1
        lne = (
            - 13-928168 * (1 - th**-1.5)
            + 34.7078238 * (1 - th**-1.25)
        )
        empa = np.exp(lne) * 0.000611657  # MPa
        e = empa * 10000.  # hPa
    if p is not None:
        e = _f(_to_hPa(p, hPa)) * e
    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa


# ---------------------------------------------------------------------
def esat_w(*args, esat=_esat, **kwargs):
    '''
    Water vapor saturation pressure over water.
    Convenience function that allows selecting actual
    formula by name

    :param esat: formula to use, possible values `None` or
        the names given under :py:meth:`meltib.humidity.set_esat`
        In case of None, the standard function is chosen.
        Defaults to `None`.
    '''
    if esat is None:
        esat = _esat
    if esat == 'goff_gratch':
        return goff_gratch_w(*args, **kwargs)
    elif esat == 'hyland_wexler':
        return hyland_wexler_w(*args, **kwargs)
    elif esat == 'iapws':
        return iapws_w(*args, **kwargs)
    elif esat == 'magnus':
        return magnus_w(*args, **kwargs)
    elif esat == 'sonntag':
        return sonntag_w(*args, **kwargs)
    elif esat == 'tetens':
        return tetens_w(*args, **kwargs)
    else:
        raise RuntimeError('esat = {} not known'.format(esat))

# ---------------------------------------------------------------------


def esat_i(*args, esat=_esat, **kwargs):
    '''
    Water vapor saturation pressure over ice.
    Convenience function that allows selecting actual
    formula by name

    :param esat: formula to use, possible values `None` or
        the names given under :py:meth:`meltib.humidity.set_esat`
        In case of None, the standard function is chosen.
        Defaults to `None`.
    '''
    if esat == 'goff_gratch':
        return goff_gratch_i(*args, **kwargs)
    elif esat == 'hyland_wexler':
        return hyland_wexler_i(*args, **kwargs)
    elif esat == 'iapws':
        return iapws_i(*args, **kwargs)
    elif esat == 'magnus':
        return magnus_i(*args, **kwargs)
    elif esat == 'sonntag':
        return sonntag_i(*args, **kwargs)
    elif esat == 'tetens':
        return tetens_i(*args, **kwargs)
    else:
        raise RuntimeError('esat = {} not known'.format(esat))

# ---------------------------------------------------------------------


def set_esat(name):
    '''
    Set standard functions `esat_w` and `esat_i`
    for the saturation pressure of water vapor.
    :param name: name of the function set (character).
    Defaults to **magnus**.

    Possible names are:
    - **goff_gratch** for `goff_gratch_w` and `goff_gratch_i`
    - **hyland_wexler** for `hyland_wexler_w` and `hyland_wexler_i`
    - **iapws** for `iapws_w` and `iapws_i`
    - **magnus** for `magnus_w` and `magnus_i`
    - **sonntag** for `sonntag_w` and `sonntag_i`
    - **tetens** for `tetens_w` and `tetens_i`
    '''
    global _esat
    if name in ['goff_gratch', 'hyland_wexler', 'magnus', 'sonntag', 'tetens']:
        _esat = name
    else:
        raise ValueError('function name {} not known'.format(name))

# ---------------------------------------------------------------------


def get_esat():
    '''
    Returns current setting of
    standard functions `esat_w` and `esat_i`
    for the saturation pressure of water vapor.

    :return: name of the function set.
    :rtype: character
    '''
    return _esat


# ---------------------------------------------------------------------
def tdew(e, p=None, hPa=False, Kelvin=True):
    '''
    Returns dewpoint, with resprect to water.

    :param e: water vaport pressuer (float)
    :param p: (optional) air pressure (float). If `p` is ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``, `p` must be supplied in hPa.
      If ``False``, `p` must be supplied in Pa.
      Defaults to ``False``.
    :param Kelvin: (optional)
      if ``True``, result is given in Kelvin.
      If ``False``, result is given in degrees Celsius.
      Defaults to ``True``.
    :return:  dewpoint
    :rtype: float
    :raises ValueError: if `e` is less equal zero.
    :raises ValueError: if `p` is less equal zero.
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, or missing

    Results are valid in the range water -45°C to +60°C for pure phase water:

    .. math::

      t = (243.12 * np.log(e / (6.112 * _f(p)))) /
          (17.62 - np.log(e / (6.112 * _f(p))))

    '''
    e = _check('e',  e, 'float', ge=0.)
    p = _check('p',  p, 'float', gt=0., none=True)
    hPa = _check('hPa', hPa, 'bool')
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)

    e = _to_hPa(e, hPa)  # hPa
    p = _to_hPa(p, hPa)  # hPa

    if np.isnan(e):
        t = np.nan
    else:
        t = (243.12 * np.log(e / (6.112 * _f(p)))) / \
            (17.62 - np.log(e / (6.112*_f(p))))  # C

    if Kelvin is False:
        return _to_C(t, Kelvin=False)  # °C
    else:
        return _to_K(t, Kelvin=False)  # K

# ---------------------------------------------------------------------


def tfrost(e, p=None, hPa=False, Kelvin=True):
    '''
    Returns frostpoint, with resprect to ice.

    :param e: water vaport pressuer (float)
    :param p: (optional) air pressure (float). If `p` is ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``, `e` and `p` must be supplied in hPa.
      If ``False``, `e` and `p` must be supplied in Pa.
      Defaults to ``False``.
    :param Kelvin: (optional)
      if ``True``, result is given in Kelvin.
      If ``False``, result is given in degrees Celsius.
      Defaults to ``True``.
    :return:  dewpoint
    :rtype: float
    :raises ValueError: if `e` is less equal zero.
    :raises ValueError: if `p` is less equal zero.
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, or missing

    Results are valid in the range water -45°C to +60°C for pure phase water:

    .. math::

      t = (272.62 * np.log(e / (6.112 * f(p)))) /
          (22.46 - np.log(e / (6.112 * f(p))))

    '''
    e = _check('e',  e, 'float', ge=0.)
    p = _check('p',  p, 'float', gt=0., none=True)
    hPa = _check('hPa', hPa, 'bool')
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)

    e = _to_hPa(e, hPa)  # Pa
    p = _to_hPa(p, hPa)  # hPa

    if np.isnan(e):
        t = np.nan
    else:
        t = (272.62 * np.log(e / (6.112 * _f(p)))) / \
            (22.46 - np.log(e / (6.112 * _f(p))))  # C

    if Kelvin is False:
        return _to_C(t, Kelvin=False)  # °C
    else:
        return _to_K(t, Kelvin=False)  # K

# ---------------------------------------------------------------------


def psychro(t, tw, Kelvin=None, p=None, hPa=False):
    '''
    Psychrometric formula for the Assmann psychrometer (liquid water).

    :param t: air or dry-bulb temperature (float)
    :param tw: wet-bulb temperature (float)
    :param Kelvin: (optional)
      if ``True``, `t` ant `tw` must be given in in Kelvin.
      Ff ``False``, `t` ant `tw` must be given in degrees Celsius.
      If missing or ``None``, unit of `t` and `tw` is autodetected.
      Defaults to ``None``.
    :param p: (optional) air pressure (float).
      if ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``, `p` must be given in Hectopascal.
      If ``False``, `p` must be given in Pascal.
      Defaults to ``True``.
    :return: water vapor pressure in Pa or hPa, depending on `hPa`
    :rtype: float
    :raises ValueError: if `p` is less equal zero.
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, ``None``,
      or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, ``None``,
      or missing

    This form of the equation was adopted by Forty-second session of the
    Executive Council (EC-XLII) in 1990 [WMO8]_:

    .. math::

      e = esat_w(t,p) - 6.53E-4 * (1 + 0.000944 * ti) * p * (t - ti)

    '''
    t = _check('t',  t, 'float', ge=-273.15)
    tw = _check('t', tw, 'float', ge=-273.15)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    p = _check('p',  p, 'float', gt=0., none=True)
    hPa = _check('hPa', hPa, 'bool')

    t = _to_C(t,  Kelvin)  # °C
    tw = _to_C(tw, Kelvin)  # °C
    if p is None:
        p1 = _to_hPa(pzero, False)  # hPa
    else:
        p1 = _to_hPa(p, hPa)  # hPa

    e = esat_w(tw, Kelvin=False, p=p, hPa=True) - 6.53E-4 * \
        (1 + 0.000944 * tw) * p1 * (t - tw)  # hPa

    if hPa is True:
        return e
    else:
        return e * 100.

# ---------------------------------------------------------------------


def psychro_ice(t, ti, Kelvin=None, p=None, hPa=False):
    '''
    Psychrometric formula for the Assmann psychrometer (ice).

    :param t: air or dry-bulb temperature (float)
    :param ti: wet-bulb temperature (float)
    :param Kelvin: (optional)
      if ``True``, `t` ant `ti` must be given in in Kelvin.
      If ``False``, `t` ant `ti` must be given in degrees Celsius.
      If missing or ``None``, unit of `t` and `ti` is autodetected.
      Defaults to ``None``.
    :param p: (optional) air pressure (float).
      if ``None`` or missing,
      the correction for moist air is omittet, i.e. :math:`f(p) = 1`.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``, `p` must be given in Hectopascal.
      If ``False``, `p` must be given in Pascal.
      Defaults to ``True``.
    :return: water vapor pressure (in Pa or hPa, depending on `hPa`
    :rtype: float
    :raises ValueError: if `p` is less equal zero.
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, ``None``,
      or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, ``None``,
      or missing

    This form of the equation was adopted by Forty-second session of the
    Executive Council (EC-XLII) in 1990 [WMO8]_:

    .. math::

      e = esat_i(t,p) - 5.75E-4 * p * (t - tw)

    '''
    t = _check('t',  t, 'float', ge=-273.15)
    ti = _check('t', ti, 'float', ge=-273.15)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    p = _check('p',  p, 'float', gt=0., none=True)
    hPa = _check('hPa', hPa, 'bool')

    t = _to_C(t,  Kelvin)  # °C
    ti = _to_C(ti, Kelvin)  # °C
    if p is None:
        p1 = _to_hPa(pzero, False)  # hPa
    else:
        p1 = _to_hPa(p, hPa)  # hPa

    e = esat_i(ti, Kelvin=False, p=p, hPa=True) - \
        5.75E-4 * p1 * (t - ti)  # hPa

    if hPa is True:
        return e  # hPa
    else:
        return 100. * e  # Pa


# ---------------------------------------------------------------------
def relhum(e, ew, percent=False):
    '''
    returns relative humidity.

    :param e: water vapor pressure (float).
      must be supplied in the same units as `ew`.
    :param ew: saturation water vapor pressure (float)
      must be supplied in the same units as `e`.
    :param percent: (optional)
      if ``True``, relative humidity is returned in percent.
      If ``False``, relative humidity is returned unitless (fraction of 1).
      Defaults to ``False``.
    :return: relative humidity (in % or unitless, depending on `percent`
    :rtype: float
    :raises ValueError: if `ew` is less equal zero.
    :raises ValueError: if `e` is less than zero.
    :raises ValueError: if `percent` is not ``True``, ``False``, or missing

    .. math::

      relhum = e / ew

    '''
    e = _check('e', e, 'float', ge=0.)
    ew = _check('ew', ew, 'float', gt=0.)
    relhum = e / ew
    if percent is True:
        return (relhum * 100.)  # %
    else:
        return (relhum)  # 1

# ---------------------------------------------------------------------


def mixr(e, p, gkg=False):
    '''
    returns approximative mixing ratio

    :param e: water vapor pressure (float).
      must be supplied in the same units as `ew`.
    :param ew: saturation water vapor pressure (float)
      must be supplied in the same units as `e`.
    :param gkg: (optional)
      if ``True``, mixr is returned in g/kg.
      If ``False``, mixr is returned in kg/kg.
      Defaults to ``False``.

    :return: relative humidity (in % or unitless, depending on `percent`
    :rtype: float
    :raises ValueError: if `e` is less equal zero.
    :raises ValueError: if `p` is less than zero.

    .. math::

      mixr = 0.62198 * e / p


    '''
    e = _check('e', e, 'float', ge=0.)
    p = _check('p', p, 'float', ge=0.)

    m = 0.62198 * e / p

    if gkg is True:
        return m * 1000.  # g/kg
    else:
        return m  # 1


# ---------------------------------------------------------------------
def inv_mixr(m, p, hPa=False, gkg=False):
    '''
    returns water vapor pressure calculated from mixing ratio

    :param m: mixing ratio (float).
    :param p: air pressure (float)
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gkg: (optional)
      if ``True``, `q`, or `m` must be supplied in g/kg.
      If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
      Defaults to ``False``.

    :return: water vapor pressure in Pa or hPa, depending on `hPa`.
    :rtype: float
    :raises ValueError: if `m` is less equal zero.
    :raises ValueError: if `q` is less than zero.

    .. math::

      e   =   m * q /  0.62198

    '''
    m = _check('m', m, 'float', ge=0.)
    p = _check('p', p, 'float', ge=0.)

    if gkg is True:
        mm = m / 1000.
    else:
        mm = m
    pp = _to_Pa(p, hPa=hPa)

    e = mm * pp / 0.62198

    if hPa is True:
        return e / 100.  # hPa
    else:
        return e  # Pa

# ---------------------------------------------------------------------


def spech(e, p, gkg=False):
    '''
    returns specific hmidity


    :param e: water vapor pressure (float).
      must be supplied in the same units as `ew`.
    :param ew: saturation water vapor pressure (float)
      must be supplied in the same units as `e`.
    :param gkg: (optional)
      if ``True``, spech is returned in g/kg.
      If ``False``, spech is returned in kg/kg.
      Defaults to ``False``.

    :return: relative humidity (in % or unitless, depending on `percent`
    :rtype: float
    :raises ValueError: if `e` is less equal zero.
    :raises ValueError: if `p` is less than zero.

    .. math::

      spech = 0.62198 * e / (e *0.622 + p)


    '''
    e = _check('e', e, 'float', ge=0.)
    p = _check('p', p, 'float', ge=0.)

    q = 0.62198 * e / (e * 0.622 + p)  # 1

    if gkg is True:
        return q * 1000.  # g/kg
    else:
        return q  # 1

# ---------------------------------------------------------------------


def inv_spech(q, p, hPa=False, gkg=False):
    '''
    returns water vapor pressure calculated from specific hmidity

    :param q: specific humidity (float).
    :param ew: saturation water vapor pressure (float)
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gkg: (optional)
      if ``True``, `q`, or `m` must be supplied in g/kg.
      If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
      Defaults to ``False``.

    :return: water vapor pressure in Pa or hPa, depending on `hPa`.
    :rtype: float
    :raises ValueError: if `e` is less equal zero.
    :raises ValueError: if `q` is less than zero.

    .. math::

      e   =   p * q / ( 0.62198 * (1 -  q))

    '''
    q = _check('q', q, 'float', ge=0.)
    p = _check('p', p, 'float', ge=0.)

    if gkg is True:
        qq = q / 1000.
    else:
        qq = q
    pp = _to_Pa(p, hPa=hPa)

    e = qq * pp / (0.62198 * (1 - qq))

    if hPa is True:
        return e / 100.  # hPa
    else:
        return e  # Pa

# ---------------------------------------------------------------------


class Humidity(object):
    '''
    The class Humidity represents a state of humid air that
    my be described and requested in any possible way

    :param t: (optional) air temperature / dry-bulb temperature (float).
    :param ew: (optional) saturation water vapor pressure (float).
     One of `t` and `ew` is required

    :param e: (optional) water vapor pressure (float).
    :param rh: (optional) relative humidity (float).
    :param td: (optional) dew point temperature (float).
    :param tf: (optional) frost point temperature (float).
    :param tw: (optional) wet-bulb temperature (float).
    :param ti: (optional) ice / frozen wet-bulb temperature (float).
    :param q: (optional) specific humidity (float).
    :param m: (optional) mixing ration (float).
    :param rhow: (optional) water vapor density in :math:`kg/m^3` (float).
    :param p: (manatory with `q` or `m`, else optional) air pressure (float).
      Defaults to 101320 Pa.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gkg: (optional)
      if ``True``, `q`, or `m` must be supplied in g/kg.
      If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
      Defaults to ``False``.
    :param percent: (optional)
      if ``True``, relative humidity is returned in percent.
      If ``False``, relative humidity is returned unitless (fraction of 1).
      Defaults to ``False``.
    :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
      or missing
    :raises ValueError: if `hPa` is not ``True``, ``False``, or missing
    :raises ValueError: if `gkg` is not ``True``, ``False``, or missing
    :raises ValueError: if `percent` is not ``True``, ``False``, or missing
    :raises ValueError:  if an invalid combination of input parameters
      is given, i.e. other than e, tf, tw and t, tw and t, q and p, m and p
    '''

    def __init__(self, t=None, p=None,
                 e=None, ew=None, rh=None, td=None, tf=None,
                 tw=None, ti=None, q=None, m=None, rhow=None,
                 Kelvin=None, hPa=False, gkg=False, percent=False):

        self.Kelvin = _check('Kelvin',  Kelvin, 'bool', none=True)
        self.hPa = _check('hPa',        hPa, 'bool')
        self.gkg = _check('gkg',        gkg, 'bool')
        self.percent = _check('percent', percent, 'bool')

        if p is None:
            self.p = pzero
        else:
            self.p = _check('p', p, 'float', gt=0.)

        loc = locals()

        tem = {x: loc[x] for x in ['t', 'ew']}
        if _only(tem, ['t']):
            if self.Kelvin:
                self.t = _check('t', t, 'float', ge=0)
            else:
                self.t = _check('t',   t, 'float', gt=-273.15)
        elif _only(tem, ['ew']):
            self.t = _check('ew', ew, 'float', gt=0.)
        else:
            raise ValueError('one of t or ew is required')

        num = {x: loc[x] for x in ['e', 'rh', 'td', 'tf',
                                   't', 'tw', 'ti', 'q', 'm', 'p', 'rhow']}
        if _only(num, ['rh', 't'], ['p']):
            if percent is True:
                rrh = _check('rh', rh, 'float', ge=0., le=100.) / 100.
            else:
                rrh = _check('rh', rh, 'float', ge=0., le=1.)
            self.e = rrh * esat_w(t, Kelvin=Kelvin, p=p, hPa=hPa)
        elif _only(num, ['e'], ['p', 't']):
            self.e = e
        elif _only(num, ['td'], ['p', 't']):
            self.e = esat_w(td, Kelvin=Kelvin, p=p, hPa=hPa)
        elif _only(num, ['tf'], ['p', 't']):
            self.e = esat_i(tf, Kelvin=Kelvin, p=p, hPa=hPa)
        elif _only(num, ['tw', 't'], ['p']):
            self.e = psychro(t, tw, Kelvin=Kelvin, p=p, hPa=hPa)
        elif _only(num, ['ti', 't'], ['p']):
            self.e = psychro_ice(t, ti, Kelvin=Kelvin, p=p, hPa=hPa)
        elif _only(num, ['q', 'p'], ['t']):
            self.e = inv_spech(q=q, p=p, hPa=hPa, gkg=gkg)
        elif _only(num, ['m', 'p'], ['t']):
            self.e = inv_mixr(m=m, p=p, hPa=hPa, gkg=gkg)
        elif _only(num, ['m', 'p'], ['t']):
            self.e = inv_mixr(m=m, p=p, hPa=hPa, gkg=gkg)
        elif _only(num, ['rhow', 't'], ['p']):
            self.e = gas_p(rhow, t, Kelvin=Kelvin, hPa=hPa, gas='water')
        else:
            raise ValueError('invalid combination of parameters: {}'.format(
                ' '.join(k for k, v in loc.items() if v is not None)))

    # ---------------------------------------------------------------------

    def rh(self, percent=None):
        '''
        returns relative humidity of the Humidity object.

        :param percent: (optional)
          if ``True``, relative humidity is returned in percent.
          If ``False``, relative humidity is returned unitless (fraction of 1).
          Defaults to object attribute `percent`.
        :return: relative humidity (in % or unitless, depending on `percent`).
        :rtype: float
        '''
        if percent is None:
            percent = self.percent
        else:
            percent = _check('percent', percent, 'bool')

        e_w = esat_w(self.t, Kelvin=self.Kelvin, p=self.p, hPa=self.hPa)
        rh = relhum(self.e, e_w)

        if percent is True:
            return rh * 100.
        else:
            return rh

    # ---------------------------------------------------------------------
    def m(self, gkg=None):
        '''
        returns mixing ratio of the Humidity object.

        :param gkg: (optional)
          if ``True``, `q`, or `m` must be supplied in g/kg.
          If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
          Defaults to object attribute `gkg`.
        :return: mixing ratio (in g/kg or unitless, depending on `gkg`).
        :rtype: float
        '''
        if gkg is None:
            gkg = self.gkg
        else:
            gkg = _check('gkg', gkg, 'bool')

        m = mixr(self.e, self.p, gkg=gkg)

        return m

    # ---------------------------------------------------------------------
    def q(self, gkg=None):
        '''
        returns specific humidity of the Humidity object.

        :param gkg: (optional)
          if ``True``, `q`, or `m` must be supplied in g/kg.
          If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
          Defaults to object attribute `gkg`.
        :return: specific humidity (in g/kg or unitless, depending on `gkg`).
        :rtype: float
        '''
        if gkg is None:
            gkg = self.gkg
        else:
            gkg = _check('gkg', gkg, 'bool')

        q = spech(self.e, self.p, gkg=gkg)

        return q

    # ---------------------------------------------------------------------
    def td(self, Kelvin=None):
        '''
        returns dew point of the Humidity object.

        :param Kelvin: (optional) if ``False``, all temperatures are assumed to
          be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: dew point (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing
        '''
        if Kelvin is None:
            Kelvin = self.Kelvin
        else:
            Kelvin = _check('Kelvin', Kelvin, 'bool')

        td = tdew(self.e, p=self.p, hPa=self.hPa, Kelvin=Kelvin)

        if Kelvin is None or Kelvin is True:
            return _to_K(td, Kelvin=Kelvin)
        else:
            return _to_C(td, Kelvin=Kelvin)

    # ---------------------------------------------------------------------
    def tf(self, Kelvin=None):
        '''
        returns frost point of the Humidity object.

        :param Kelvin: (optional)
          if ``False``, all temperatures are assumed to be Kelvin.
          If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: frost point (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing
        '''
        if Kelvin is None:
            Kelvin = self.Kelvin
        else:
            Kelvin = _check('Kelvin', Kelvin, 'bool')

        tf = tfrost(self.e, p=self.p, hPa=self.hPa, Kelvin=Kelvin)

        if Kelvin is None or Kelvin is True:
            return _to_K(tf, Kelvin=Kelvin)
        else:
            return _to_C(tf, Kelvin=Kelvin)

    # ---------------------------------------------------------------------
    def tw(self, Kelvin=None):
        '''
        returns wet-bulb temperature of the Humidity object.

        :param Kelvin: (optional)
          if ``False``, all temperatures are assumed to be Kelvin.
          If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: wet-bulb temperatur (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing

        Since there is no closed analytical solution, we use the
        iterative method of [MAR2009]_,
        adapted for use with the Magnus formula :eq:`magnus` for
        saturation water vapor from [WMO8]_:

        1) set inital value

        .. math::
          tw_0 = t - 5

        2) calculate next value:

        .. math::
          tw' &= tl - \\frac{ ew(tw_i) - e}
          { 6.53E-4 * (1 + 0.000944 * tw_i) * p} \\\\
          tw_{i+1} &= tw_i + 0.5 * (tw' - tw_i)

        3) iterate until:

        .. math::
          \\left| tw_{i+1} - tw_i \\right| < 0.01K

        If the iteration does not converge within 100 iterations,
        ``np.nan`` is returned

        '''
        if Kelvin is None:
            Kelvin = self.Kelvin
        else:
            Kelvin = _check('Kelvin', Kelvin, 'bool')

        tl = _to_C(self.t, Kelvin=self.Kelvin)  # °C
        pp = _to_hPa(self.p, hPa=self.hPa)  # hPa
        ee = _to_hPa(self.e, hPa=self.hPa)  # hPa

        tw = tl - 5  # °C
        for i in range(100):
            twl = tw
            twn = tl - (esat_w(tw, Kelvin=False, hPa=True) - ee) / \
                (6.53E-4 * (1 + 0.000944 * tw) * pp)  # °C
            tw = tw + 0.5 * (twn - tw)
            if abs(tw - twl) <= 0.01:
                break
        else:
            tw = np.nan

        if Kelvin is False:
            return tw
        else:
            return _to_K(tw, Kelvin=False)

    # ---------------------------------------------------------------------
    def ti(self, Kelvin=None):
        '''
        returns ice / frozen wet-bulb temperature of the Humidity object.

        :param Kelvin: (optional) if ``False``, all temperatures are assumed to
          be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: ice / frozen wet-bulb temperatur (in K or °C,
          depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing

        Since there is no closed analytical solution, we use the
        iterative method of [MAR2009]_,
        adapted for use with the Magnus formula :eq:`magnice` for
        saturation water vapor from [WMO8]_:

        1) set inital value

        .. math::
          ti_0 = t - 5

        2) calculate next value:

        .. math::
          ti' &= tl - \\frac{ ew(ti_i) - e}{ 5.75E-4 * p } \\\\
          ti_{i+1} &= ti_i + 0.5 * (ti' - ti_i)

        3) iterate until:

        .. math::
          \\left| ti_{i+1} - ti_i \\right| < 0.01K

        If the iteration does not converge within 100 iterations,
        ``np.nan`` is returned
        '''
        if Kelvin is None:
            Kelvin = self.Kelvin
        else:
            Kelvin = _check('Kelvin', Kelvin, 'bool')

        tl = _to_C(self.t, Kelvin=self.Kelvin)  # °C
        pp = _to_hPa(self.p, hPa=self.hPa)  # hPa
        ee = _to_hPa(self.e, hPa=self.hPa)  # hPa

        ti = tl - 5  # °C
        for i in range(100):
            til = ti
            tin = tl - (esat_i(ti, Kelvin=False, hPa=True) - ee) / \
                (5.75E-4 * pp)  # °C
            ti = ti + 0.5 * (tin - ti)
            if abs(ti - til) <= 0.01:
                break
        else:
            ti = np.nan

        if Kelvin is False:
            return ti
        else:
            return _to_K(ti, Kelvin=False)

    # ---------------------------------------------------------------------
    def virt_inc(self):
        '''
        returns virtual temperature increment of the Humidity object.

        :return: virtual temperature increment (in K).
        :rtype: float

        '''
        T = _to_K(self.t, Kelvin=self.Kelvin)
        qq = spech(self.e, self.p, gkg=False)

        vi = T * 0.61 * qq

        return vi

    # ---------------------------------------------------------------------
    def tvirt(self, Kelvin=None):
        '''
        calculates virtual temperature of the Humidity object.

        :param Kelvin: (optional) if ``False``, all temperatures are assumed to
          be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: virtual temperature (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing
        '''
        if Kelvin is None:
            Kelvin = self.Kelvin
        else:
            Kelvin = _check('Kelvin', Kelvin, 'bool')

        vi = self.virt_inc()

        T = _to_K(self.t, Kelvin=self.Kelvin)  # K
        Tv = T + vi  # K

        if Kelvin is False:
            return _to_C(Tv, Kelvin=True)
        else:
            return _to_K(Tv, Kelvin=True)

    # ---------------------------------------------------------------------
    def tequi(self, Kelvin=None):
        '''
        calculates equivalent temperature of the Humidity object.

        :param Kelvin: (optional)
          if ``False``, all temperatures are assumed to be Kelvin.
          If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: equivalent temperature (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing

        equivalent temperature is the temperature of an air parcel from which
        all the water vapor has been removed by an adiabatic process:

        .. math:: T_{e} \\approx T + \\frac{L_v}{c_p} m

        '''
        mm = self.m(gkg=False)

        T = _to_K(self.t, Kelvin=self.Kelvin)

        Te = T + mm * Lv / cp

        if Kelvin is False:
            return _to_C(Te, Kelvin=True)
        else:
            return _to_K(Te, Kelvin=True)

    # ---------------------------------------------------------------------
    def tequipot(self, Kelvin=None):
        '''
        calculates equipotential temperature of the Humidity object.

        :param Kelvin: (optional) if ``False``, all temperatures are assumed to
          be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
          If missing of ``None``, unit  temperatures are autodetected.
          Defaults to ``None``.
        :return: equipotential temperature (in K or °C, depending on `Kelvin`).
        :rtype: float
        :raises ValueError: if `Kelvin` is not ``True``, ``False``, None,
          or missing

        equipotential temperature is the temperature of an air parcel
        from which all the water vapor has been removed
        by an adiabatic process [Stu1988]_:

        .. math:: \\Theta_{e} = T_e
                  \\left( \\frac{p_0}{p} \\right) ^{\\frac{R}{c_p}}
                  \\approx \\left( T + \\frac{L_v}{c_p} m \\right)
                  \\left( \\frac{p_0}{p} \\right)^{\\frac{R}{c_p}}

        '''
        te = self.tequi()

        tep = Tpot(te, p=self.p, Kelvin=self.Kelvin, hPa=self.hPa)

        return tep

    # ---------------------------------------------------------------------
    def rhow(self):
        '''
        calculates water vapor density of the Humidity object.

        :return: density (in :math:`kg/m^3` ).
        :rtype: float

        '''
        rw = gas_rho(self.e, self.t, gas='water',
                     Kelvin=self.Kelvin, hPa=self.hPa)

        return rw

# ---------------------------------------------------------------------


def __init__():
    print('tralala')
    set_esat('magnus')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Temperature calculations an conversions
'''

import numpy as np

from ._utils import _check
from .constants import R, Tzero, cp
from .pressure import _to_Pa

"""Reference pressure for calculation of potential temperatures"""
POTENTIAL_REFERENCE_PRESSURE = 100000.  # Pa

# ---------------------------------------------------------------------


def _to_K(t, Kelvin=None):
    """
    Returns the given Temperature in Kelvin

    :param t: Temperature
    :param Kelvin: (optional) if ``False``, unit of t is assumed to
      be Kelvin. If ``False``, unit of t is assumed to be Celsius.
      If missing of ``None``, unit of t is autodetected. Defaults to
      ``None``.
    :return: temperature in K. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float
    :raises ValueError: if Kelvin is not ``True``, ``False``, or ``None``

    """
    try:
        _check('t', t, 'float', nan=True)
    except ValueError:
        return np.nan
    if Kelvin is None:
        # autodetect
        if np.all(np.less(t, 150.)):
            t = t + Tzero
    elif Kelvin is False:
        t = t + Tzero
    elif Kelvin is True:
        pass
    else:
        raise ValueError('invalid value for "Kelvin": %s' % format(Kelvin))
    return t

# ---------------------------------------------------------------------


def _to_C(t, Kelvin=None):
    '''
    Returns the given Temperature in Celsius

    :param t: Temperature
    :param Kelvin: (optional) if ``False``, unit of t is assumed to
      be Kelvin. If ``False``, unit of t is assumed to be Celsius.
      If missing of ``None``, unit of t is autodetected. Defaults to
      ``None``.
    :return: temperature in °C. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float
    :raises ValueError: if Kelvin is not ``True``, ``False``, or ``None``

    '''
    try:
        _check('t', t, 'float', nan=True)
    except ValueError:
        return np.nan
    if Kelvin is None:
        # autodetect
        if t >= 150.:
            t = t - Tzero
    elif Kelvin is False:
        pass
    elif Kelvin is True:
        t = t - Tzero
    else:
        raise ValueError('invalid value for "Kelvin": %s' % format(Kelvin))
    return t

# ---------------------------------------------------------------------


def KtoC(t):
    '''
    Converts Temperature from Kelvin to Celsius

    :param t: Temperature K
    :return: temperature in °C. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float

    '''
    return _to_C(t, Kelvin=True)

# ---------------------------------------------------------------------


def CtoK(t):
    '''
    Converts Temperature from Celsius to Kelvin

    :param t: Temperature in °C
    :return: temperature in K. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float

    '''
    return _to_K(t, Kelvin=False)

# ---------------------------------------------------------------------


def FtoC(t):
    '''
    Converts Temperature from Fahrenheit to Celsius

    :param t: Temperature in °F (float)
    :return: temperature in °C.
    :rtype: float

    Conversion:

    .. math:: T_C = (T_F + 32.0) \\cdot \\frac{5}{9}

    '''
    try:
        _check('t', t, 'float')
    except ValueError:
        return np.nan
    return (t - 32.0) * 5./9.


# ---------------------------------------------------------------------
def CtoF(t):
    '''
    Converts Temperature from Celsius to Fahrenheit

    :param t: Temperature in °C (float)
    :return: temperature in °F.
    :rtype: float

    Conversion:

    .. math:: T_F = T_C \\cdot 1.80 + 32

    '''
    try:
        _check('t', t, 'float')
    except ValueError:
        return np.nan
    return t*1.80 + 32


# ---------------------------------------------------------------------
def Tpot(t, p, Kelvin=None, hPa=False, pref=POTENTIAL_REFERENCE_PRESSURE):
    '''
    Converts Temperature from Celsius to Kelvin

    :param t: Temperature in °C
    :param p: (optional) air pressure (float).
    :param Kelvin: (optional) if ``False``, unit of t is assumed to
      be Kelvin. If ``False``, unit of t is assumed to be Celsius.
      If missing of ``None``, unit of t is autodetected. Defaults to
      ``None``.
    :param hPa: (optional)
      if ``False``, `p` must be supplied in Pa .
      If ``True``, `p` must be supplied in hPa.
      Defaults to ``False``
    :param pref: (optional)
      Defaults to 100 000 Pa.
    :return: temperature in K. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float

    '''
    T = _to_K(_check('t', t, 'float', nan=True), Kelvin=Kelvin)
    pp = _to_Pa(_check('p', p, 'float', gt=0., nan=True), hPa=hPa)

    Theta = T * (pref / pp)**(R/cp)

    if Kelvin is False:
        return _to_C(Theta, Kelvin=True)
    return _to_K(Theta, Kelvin=True)

# ---------------------------------------------------------------------


def inv_Tpot(t, p, Kelvin=None, hPa=False, pref=POTENTIAL_REFERENCE_PRESSURE):
    '''
    Converts Temperature from Celsius to Kelvin

    :param t: Temperature in °C
    :param p: (optional) air pressure (float).
    :param Kelvin: (optional) if ``False``, unit of t is assumed to
      be Kelvin. If ``False``, unit of t is assumed to be Celsius.
      If missing of ``None``, unit of t is autodetected. Defaults to
      ``None``.
    :param hPa: (optional)
      if ``False``, `p` must be supplied in Pa .
      If ``True``, `p` must be supplied in hPa.
      Defaults to ``False``
    :param pref: (optional)
      Defaults to 100 000 Pa.
    :return: temperature in K. If t is not a valid number, ``np.nan``
      is returned.
    :rtype: float

    '''
    Theta = _to_K(_check('t', t, 'float', nan=True), Kelvin=Kelvin)
    pp = _to_Pa(_check('p', p, 'float', gt=0., nan=True), hPa=hPa)

    T = Theta / (pref / pp)**(R/cp)

    if Kelvin is False:
        return _to_C(T, Kelvin=True)
    return _to_K(T, Kelvin=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 18:03:56 2019

@author: druee
"""

import unittest

import meteolib as m


class Test_to_K(unittest.TestCase):
    def test_OC_auto(self):
        """
        0°C -> 273.15 K
        """
        result = m.temperature._to_K(0.)
        self.assertEqual(result, m.constants.Tzero)

    def test_273C_auto(self):
        """
        273.15 K -> 273.15 K
        """
        result = m.temperature._to_K(m.constants.Tzero)
        self.assertEqual(result, m.constants.Tzero)

    def test_OC_C(self):
        """
        0°C -> 273.15 K
        """
        result = m.temperature._to_K(0., False)
        self.assertEqual(result, m.constants.Tzero)

    def test_273K_K(self):
        """
        273.15 K -> 273.15 K
        """
        result = m.temperature._to_K(m.constants.Tzero, True)
        self.assertEqual(result, m.constants.Tzero)


class Test_to_C(unittest.TestCase):
    def test_OC_auto(self):
        """
        0°C -> 0°C
        """
        result = m.temperature._to_C(0.)
        self.assertEqual(result, 0.)

    def test_273K_auto(self):
        """
        273.15 K -> 0°C
        """
        result = m.temperature._to_C(m.constants.Tzero)
        self.assertEqual(result, 0.)

    def test_OC_C(self):
        """
        0°C -> 0°C
        """
        result = m.temperature._to_C(0., False)
        self.assertEqual(result, 0.)

    def test_273K_K(self):
        """
        273.15 K -> 273.15 K
        """
        result = m.temperature._to_C(m.constants.Tzero, True)
        self.assertEqual(result, 0.)


class Test_KtoC(unittest.TestCase):
    def test_273K(self):
        """
        298.15 K -> 25°C
        """
        result = m.temperature.KtoC(298.15)
        self.assertEqual(result, 25.)


class Test_CtoK(unittest.TestCase):
    def test_273K(self):
        """
        273.15 K -> 0°C
        """
        result = m.temperature.CtoK(25.)
        self.assertEqual(result, 298.15)


class Test_FtoC(unittest.TestCase):
    def test_0F(self):
        """
        0 F = -17.77C
        """
        result = m.temperature.FtoC(0)
        self.assertAlmostEqual(result, -17.77, delta=0.01)

    def test_0C(self):
        """
        0 C = 32.0F
        """
        result = m.temperature.FtoC(32.0)
        self.assertAlmostEqual(result, 0., delta=0.01)


class Test_Tpot(unittest.TestCase):
    def test_pref_auto(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.Tpot(
            273.15, 100000.), 273.15, delta=0.01)
        self.assertAlmostEqual(m.temperature.Tpot(
            0., 100000.), 273.15, delta=0.01)

    def test_pref_K(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.Tpot(
            273.15, 100000., Kelvin=True), 273.15, delta=0.01)

    def test_pref_C(self):
        """
        0°C, 100000 Pa -> 0 °C
        """
        self.assertAlmostEqual(m.temperature.Tpot(
            0., 100000., Kelvin=True), 0., delta=0.01)

    def test_pref_hPa(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.Tpot(
            273.15, 1000., hPa=True), 273.15, delta=0.01)

    def test_750_auto(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.Tpot(
            293.15, 75000.), 318.28, delta=0.03)
        self.assertAlmostEqual(m.temperature.Tpot(
            20., 75000.), 318.28, delta=0.03)


class Test_inv_Tpot(unittest.TestCase):
    def test_pref_auto(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            273.15, 100000.), 273.15, delta=0.01)
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            0., 100000.), 273.15, delta=0.01)

    def test_pref_K(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            273.15, 100000., Kelvin=True), 273.15, delta=0.01)

    def test_pref_C(self):
        """
        0°C, 100000 Pa -> 0 °C
        """
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            0., 100000., Kelvin=True), 0., delta=0.01)

    def test_pref_hPa(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            273.15, 1000., hPa=True), 273.15, delta=0.01)

    def test_750_auto(self):
        """
        273.15 K, 100000 Pa -> 273.15 K
        """
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            318.28, 75000.), 293.15, delta=0.03)
        self.assertAlmostEqual(m.temperature.inv_Tpot(
            45.13, 75000.), 293.15, delta=0.03)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
physical constants

'''

#: Stefan-Boltzmann constant (:math:`W m^{-2} K^{-4}`)
#: of the identically named physical law that relatet
#: electromagnetig radion intensity to black-body temperature.
#: :math:`P = \sigma T^4`
#: Recommended value Adopted from [CODATA]_.
sigma = 5.570367E-8

#: Absolute temperature (:math:`K`)
#: of the normal ice point
#: Offset between Kelvin and Celsius Tmeperature scales.
#: Adopted by WMO in [WMO8]_.
Tzero = 273.15

#: Absolute temperature (:math:`K`)
#: of the triple point of water
#: by definition of [ITS90]_.
#: Adopted by WMO in [WMO8]_.
Ttriple = 273.16


#:  Standard acceleration of gravity (:math:`m s^{-2}`).
#:  Adopted by WMO in [WMO8]_.
gn = 9.806

#: Density of mercury (:math:`kg m^{–3}`) at 0 °C.
#: Adopted by WMO in [WMO8]_.
rhoHg = 1.35951E4

#: Reference pressure (:math:`Pa`).
#: One of the typical references levels --
#: Mean Sea Level (MSL), station altitude or the 1013.2 hPa plane.
#: -- adopted by WMO in [WMO8]_.
pzero = 101320.

# ----------------------------------------------------------------------
# WMO 180, Table 4.1: THERMODYNAMIC CONSTANTS AND FUNCTIONS
# (except constants above)

# 1. Basic constants

#: Apparent molecular weight of dry air (:math:`kg mol^{-1}`).
#: Adopted by WMO in [WMO188]_.
M = 28.9644E-3

#: Gas constant for 1 mole of ideal gas (:math:`J mol^{-1}`)
#: Adopted by WMO in [WMO188]_.
Rstar = 8.31432

#: Gas constant for dry air (:math:`J kg^-1 K^-1`).
#: Adopted by WMO in [WMO188]_.
R = 287.05

#: Molecular weight of water vapour (:math:`kg mol^{-1}`).
#: Adopted by WMO in [WMO188]_.
Mw = 18.0153E-3

#: Gas constant for water vapour (:math:`J kg^-1 K^-1`).
#: Adopted by WMO in [WMO188]_.
Rw = 461.51

# 2. Specific heat capacities

#: Specific heat capacity of **dry air**
#: at constant pressure (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.cpt`
cp = 1005.

#: Specific heat capacity of **dry air**
#: at constant volume (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
cv = 718.

#: Specific heat capacity of **liquid water** (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.cwt`
cw = 4179.

#: Specific heat capacity of frozen water / **ice**  (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.cit`
ci = 2090.

#: Specific heat capacity of **water vapor**
#: at constant pressure (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.cpvt`
cpv = 1850.

#: Specific heat capacity of **water vapor**
#: at constant volume (:math:`J kg^-1 K^-1`).
#: Recommended by WMO in [WMO188]_.
cvv = 1390.


# 3. Heats of transformation of phase of water

#: Specific heat of sublimation (:math:`J kg^-1`).
#: Recommended by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.Lst`
Ls = 2.835E6

#: Specific heat of vaporization / evaportaion (:math:`J kg^-1`).
#: Mean value named by WMO in [WMO188]_.
#: For temperature-dependent formulation see :py:meth:`meteolib.thermodyn.Lvt`
Lv = 2.501E6

# ----------------------------------------------------------------------
# other constants

#: Apparent molecular weight of carbon dioxide (:math:`kg mol^{-1}`).
#: Adopted by NIST [NIST]_.
Mco2 = 44.0095E-3

#: Gas constant for carbon dioxide (:math:`J kg^-1 K^-1`).
#: calculated from `Rstar` and `Mco2`.
Rco2 = Rstar/Mco2

#: von-Karman constant. Re-evaluated value by [Hog1985]_
kappa = 0.40

#: Solar constant (:math:`W m^{-2}`)
#: a synonym for mean extra-terrestrial solar irradiance.
#: said to be defined by WMO and used in WMO documents
#: and many research papers, no immediate source.
solar = 1367.

#: Eart's nominal radius (m)
#: Defined in [ISO2533]_
r_earth = 6356766



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 18:39:22 2019

@author: druee
"""

import unittest

import meteolib as m


class Test__check(unittest.TestCase):
    """
    def _check(name,par,kind,lt=None,le=None,ne=None,ge=None,gt=None)
    """

    def test_1_float(self):
        result = m._utils._check('x', 1., 'float')
        self.assertEqual(result, 1.)

    def test_1_int(self):
        result = m._utils._check('x', 1, 'int')
        self.assertEqual(result, 1)

    def test_wrong_kind(self):
        self.assertRaises(ValueError, m._utils._check, 'x', 1, 'blah')

    def test_range_bad(self):
        with self.assertRaises(ValueError):
            m._utils._check('x', 1., 'float', lt=1.)
            m._utils._check('x', 1., 'float', le=0.)
            m._utils._check('x', 1., 'float', ne=1.)
            m._utils._check('x', 1., 'float', ge=1.)
            m._utils._check('x', 1., 'float', gt=2.)

    def test_range_good(self):
        self.assertEqual(m._utils._check('x', 1., 'float', lt=2.), 1.)
        self.assertEqual(m._utils._check('x', 1., 'float', le=1.), 1.)
        self.assertEqual(m._utils._check('x', 1., 'float', ge=1.), 1.)
        self.assertEqual(m._utils._check('x', 1., 'float', gt=0.), 1.)


class Test__only(unittest.TestCase):
    '''
    def _only(pars,sel):
    '''

    def test_types(self):
        self.assertRaises(TypeError, m._utils._only, {
                          'a': None, 'b': None, 'c': None, 'd': None}, 1)
        self.assertRaises(TypeError, m._utils._only, 1, ['b'])

    def test_true_list(self):
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': None, 'd': None}, ['b']))
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': 1, 'd': None}, ['b', 'c']))

    def test_true_str(self):
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': None, 'd': None}, 'b'))

    def test_false_list(self):
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': None, 'd': None}, ['c']))
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': 1, 'd': None}, ['a', 'd']))
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': 1, 'd': None}, ['a', 'b']))
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': 1, 'd': None}, ['b']))

    def test_false_str(self):
        self.assertRaises(TypeError, m._utils._only(
            {'a': None, 'b': 1, 'c': None, 'd': None}, 'd'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 18:39:22 2019

@author: druee
"""

import unittest

import meteolib as m


class Test_dir2uv(unittest.TestCase):
    """
    def dir2uv(ff,dd)
    """

    def test_angles(self):
        res, ref = m.wind.dir2uv(1.,  0.), (0., -1.)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)
        res, ref = m.wind.dir2uv(1., 180.), (0.,  1.)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)
        res, ref = m.wind.dir2uv(1., 270.), (1.,  0.)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)
        res, ref = m.wind.dir2uv(1., 360.), (0., -1.)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)

    def test_speeds(self):
        with self.assertRaises(ValueError):
            m.wind.dir2uv(-1., 0.)
        res, ref = m.wind.dir2uv(2., 270.), (2.,  0.)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)
        res, ref = m.wind.dir2uv(2., 45.), (-1.4142,  -1.4142)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)

    def test_shape(self):
        with self.assertRaises(ValueError):
            m.wind.dir2uv([0], [0, 0])
        res, ref = m.wind.dir2uv([2.], [45.]), (-1.4142,  -1.4142)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=4)
        res, ref = m.wind.dir2uv(
            [2., 2.], [45., 45.]), ([-1.4142,  -1.4142], [-1.4142,  -1.4142])
        for s, f in zip(res, ref):
            for a, b in zip(s, f):
                self.assertAlmostEqual(a, b, places=4)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Aug  7 18:39:22 2019

@author: druee
"""

import datetime as dt
import unittest

import meteolib as m

import pytz


CET = dt.timezone(dt.timedelta(hours=1))
UTC = dt.timezone(dt.timedelta(hours=0))

MST = pytz.timezone('MST')


class Test_sunriseset(unittest.TestCase):
    """
    sun_rise_transit_set(time: dt.datetime, lat, lon, ele=0, pp=None, tk=None)
    """
    def test_spa_rts_values(self):
        res, ref = m.radiation.spa_rise_transit_set(
            dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET), 49.6, 6.9), (
            8.477979163121788, 12.624062013177186, 16.77403260994837)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, places=3)

    def test_fast_rts_values(self):
        res, ref = m.radiation.fast_rise_transit_set(
            dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET), 49.6, 6.9), (
            8.477979163121788, 12.624062013177186, 16.77403260994837)
        for s, f in zip(res, ref):
            self.assertAlmostEqual(s, f, delta=0.25)

    def test_spa_pos_values(self):
        # reference values calculated by NREL SOLPOS Calculator
        # https://midcdmz.nrel.gov/solpos/solpos.html
        fun = m.radiation.spa_sun_position
        tuples = [
            # (mon,day, hr, lat,lon, ele, azi)
            # equator Jan
            (1, 1,  9, 0, 0, 39.8611, 120.5933),
            (1, 1, 12, 0, 0, 66.9985, 177.9096),
            # equator July
            (7, 1,  9, 0, 0, 39.8611, 59.3124),
            (7, 1, 12, 0, 0, 66.9985, 2.519),
            # 50N July
            (7, 1, 9, 50, 0, 45.3667, 109.7713),
            (7, 1, 12, 50, 0, 63.0830, 178.0475),
            # 50N 90E July
            (7, 1,  9, 50, 90, 46.5148, 248.3167),
            (7, 1, 12, 50, 90, 18.0823, 284.6427),
            # pole July
            (7, 1,  9, 89, 0, 23.7953, 133.7327),
            (7, 1, 12, 89, 0, 24.0934, 179.0427),
        ]
        for mon, day, hr, lat, lon, ele, azi in tuples:
            res = fun(
                dt.datetime(2010, mon, day, hr, 0, 0, tzinfo=UTC),
                lat, lon
            )
            ref = (ele, azi)
            for s, f in zip(res, ref):
                self.assertAlmostEqual(s, f, delta=0.5)

    def test_fast_pos_values(self):
        # reference values calculated by NREL SOLPOS Calculator
        # https://midcdmz.nrel.gov/solpos/solpos.html
        fun = m.radiation.fast_sun_position
        tuples = [
            # (mon,day, hr, lat,lon, ele, azi)
            # equator Jan
            (1, 1,  9, 0, 0, 39.8611, 120.5933),
            (1, 1, 12, 0, 0, 66.9985, 177.9096),
            # equator July
            (7, 1,  9, 0, 0, 39.8611, 59.3124),
            (7, 1, 12, 0, 0, 66.9985, 2.519),
            # 50N July
            (7, 1, 9, 50, 0, 45.3667, 109.7713),
            (7, 1, 12, 50, 0, 63.0830, 178.0475),
            # 50N 90E July
            (7, 1,  9, 50, 90, 46.5148, 248.3167),
            (7, 1, 12, 50, 90, 18.0823, 284.6427),
            # pole July
            (7, 1,  9, 89, 0, 23.7953, 133.7327),
            (7, 1, 12, 89, 0, 24.0934, 179.0427),
        ]
        for mon, day, hr, lat, lon, ele, azi in tuples:
            res = fun(
                dt.datetime(2010, mon, day, hr, 0, 0, tzinfo=UTC),
                lat, lon
            )
            ref = (ele, azi)
            for s, f in zip(res, ref):
                self.assertAlmostEqual(s, f, delta=0.5)

    def test_spa_rts_wrongargs(self):
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=980, tk=-1)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100, 6.9, 252, pp=980, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                -100, 5.9, 252, pp=980, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, -181, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 361, 252, pp=-1, tk=288)

    def test_fast_rts_wrongargs(self):
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(2110, 1, 1, 14, 0, 0, tzinfo=CET),
                49.6, 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(1890, 1, 1, 14, 0, 0, tzinfo=CET),
                49.6, 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(1890, 1, 1, 14, 0, 0, tzinfo=CET),
                100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                -100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, -181)
        with self.assertRaises(ValueError):
            m.radiation.fast_rise_transit_set(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100, 361)

    def test_spa_pos_wrongargs(self):
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=980, tk=-1)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 6.9, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100, 6.9, 252, pp=980, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                -100, 5.9, 252, pp=980, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, -181, 252, pp=-1, tk=288)
        with self.assertRaises(ValueError):
            m.radiation.spa_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, 361, 252, pp=-1, tk=288)

    def test_fast_pos_wrongargs(self):
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(2110, 1, 1, 14, 0, 0, tzinfo=CET),
                49.6, 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(1890, 1, 1, 14, 0, 0, tzinfo=CET),
                49.6, 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(1890, 1, 1, 14, 0, 0, tzinfo=CET),
                100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                -100., 6.9)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                49.6, -181)
        with self.assertRaises(ValueError):
            m.radiation.fast_sun_position(
                dt.datetime(2021, 1, 4, 14, 0, 0, tzinfo=CET),
                100, 361)


class Test_irradiance(unittest.TestCase):

    def test_clear_sky_direct_normal_values(self):
        # 11:47 UTC = 12:00 local in Jan
        # 12:01 UTC = 12:00 local in Jul
        # reference values: [ASH1997]_ table 18 page 29.32
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 1, 21, 11, 47, 0, tzinfo=UTC),
                40.0, 0), 926, delta=2)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 1, 21, 5, 47, 0, tzinfo=UTC),
                40.0, 0), 0, delta=1)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 12, 1, 0, tzinfo=UTC),
                40.0, 0), 879, delta=1)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 9, 1, 0, tzinfo=UTC),
                40.0, 0), 829, delta=2)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 6, 1, 0, tzinfo=UTC),
                40.0, 0), 488, delta=3)

        # reference values [ASH1997]_ table 21 page 29.35
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 1, 21, 11, 47, 0, tzinfo=UTC),
                64.0, 0), 316, delta=4)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 1, 21, 5, 47, 0, tzinfo=UTC),
                64.0, 0), 0, delta=2)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 12, 1, 0, tzinfo=UTC),
                64.0, 0), 831, delta=2)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 9, 1, 0, tzinfo=UTC),
                64.0, 0), 791, delta=2)
        self.assertAlmostEqual(
            m.radiation.clear_sky_direct_normal(
                dt.datetime(2020, 6, 21, 6, 1, 0, tzinfo=UTC),
                64.0, 0), 613, delta=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''Thermodynamics laws and equations

'''
import numpy as np

from ._utils import _check
from .constants import R, Rco2, Rw
from .pressure import _to_hPa
from .temperature import _to_C, _to_K


def gas_p(rho, T, q=None, Kelvin=None, hPa=False, gkg=False, gas='air'):
    r'''
    ideal gas equation, calculate gas pressure

    :param rho: density of gas in :math:`kg/m^3` .
    :param T:  gas temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin` .
    :param rho: density of gas (in :math:`kg/m^3` .
    :param gas: kind of gas.
      One of `dry` (dry air), `moist` (humid air), `water` (water vapor),
      or `co2` (carbon dioxide).
      Defaults to `dry`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected.
      Defaults to ``None``.
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gkg: (optional)
      if ``True``, `q`, or `m` must be supplied in g/kg.
      If ``False``, `q`, or `m` must be supplied unitless, i.e. in kg/kg.
      Defaults to ``False``.

    :return: air pressure (in hPa or Pa, depending on `hPa`).
    :rtype: float

    '''
    if gas in ['dry', 'moist']:
        Rgas = R
    elif gas in ['water']:
        Rgas = Rw
    elif gas in ['co2']:
        Rgas = Rco2
    else:
        raise ValueError('unknow value for "gas": {}'.format(gas))
    T = _to_K(T, Kelvin)
    if gas in ['dry', 'water', 'co2']:
        Tgas = T
    elif gas == 'moist':
        #
        # virtual Temperature
        # hardcoded bevcause cannot be impoprtet mutually
        if gkg is True:
            Tgas = T * (1 + 0.00061*q)
        else:
            Tgas = T * (1 + 0.61*q)

    p = rho * Rgas * Tgas

    if hPa is True:
        return _to_hPa(p, hPa=False)
    else:
        return p


def gas_rho(p, T, q=None, Kelvin=None, hPa=False, gkg=False, gas='air'):
    r'''
    ideal gas equation, calculate gas density

    :param p: air pressure (in hPa or Pa, depending on `hPa`).
    :param T:  gas temperature in :math:`K` or :math:`^\circ C` ,
      depending on `Kelvin` .
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gas: kind of gas.
      One of `dry` (dry air), `moist` (humid air), `water` (water vapor),
      or `co2` (carbon dioxide).
      Defaults to `dry`.

    :return: density of gas in :math:`kg/m^3`.
    :rtype: float

    '''
    if gas in ['air', 'dry', 'moist']:
        Rgas = R
    elif gas in ['water']:
        Rgas = Rw
    elif gas in ['co2']:
        Rgas = Rco2
    else:
        raise ValueError('unknow value for "gas": {}'.format(gas))
    T = _to_K(T, Kelvin)
    if gas in ['air', 'dry', 'water', 'co2']:
        Tgas = T
    elif gas == 'moist':
        #
        # virtual Temperature
        # hardcoded bevcause cannot be impoprtet mutually
        if gkg is True:
            Tgas = T * (1 + 0.00061*q)
        else:
            Tgas = T * (1 + 0.61*q)
    if hPa is True:
        pgas = p*100.
    else:
        pgas = p

    rho = pgas / (Rgas * Tgas)

    return rho


def gas_t(p, rho, q=None, Kelvin=None, hPa=False, gas='air'):
    r'''
    ideal gas equation, calculate gas temperature

    :param p: air pressure (in hPa or Pa, depending on `hPa`).
    :param rho: density of gas in :math:`kg/m^3` .
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.
    :param hPa: (optional)
      if ``True``,  `p`, `e`, or `ew` must be supplied in hPa.
      If ``False``, `p`, `e`, or `ew` must be supplied in Pa.
      Defaults to ``False``.
    :param gas: kind of gas.
      One of `dry` (dry air), `moist` (humid air), `water` (water vapor),
      or `co2` (carbon dioxide).
      Defaults to `dry`.

    :return: gas temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin` .
    :rtype: float

    '''
    if gas in ['air', 'dry', 'moist']:
        Rgas = R
    elif gas in ['water']:
        Rgas = Rw
    elif gas in ['co2']:
        Rgas = Rco2
    else:
        raise ValueError('unknow value for "gas": {}'.format(gas))

    Tgas = p / (Rgas * rho)

    if gas in ['air', 'dry', 'water', 'co2']:
        T = Tgas
    elif gas == 'moist':
        #    Tgas = tvirt_inv(Tgas,q,Kelvin=True,gkg=gkg)
        raise RuntimeError('not yet implemented')

    if Kelvin is True:
        return T
    else:
        return _to_C(T, Kelvin=True)

# ----------------------------------------------------------------------
# not so constant constants
# ----------------------------------------------------------------------


def cpt(t, Kelvin=None):
    r'''
    Specific heat capacity of dry air, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin` .
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.

    :return: specific heat capacity of **dry air** at constant pressure
      (in :math:`J kg^-1 K^-1`).
    :rtype: float

    .. math::

      c_{p} ~=~ 1005.60 + 0.017211 T + 0.000392 T^2

    valid for temperature t between –40°C and 40°C and
    for barometric pressures near one atmosphere [And2006]_.
    Minimum value of 1005.41:math:`J kg^-1 K^-1` at 21.95°C.

    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    #
    th = _to_C(t, Kelvin=Kelvin)
    cp = 1005.60 + 0.017211 * th + 0.000392 * (th**2)
    # equation (A20) in Garrat, J.R., 1992. 'The Atmospheric Boundary
    #     layer', Cambridge University Press:
    # cp = 1005.16 + 0.013763 * th + 2.97265E-4 * (th**2)
    return cp

# ----------------------------------------------------------------------


def cpvt(t, Kelvin=None):
    r'''
    Specific heat capacity of water vapor, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.

    :return: specific heat capacity of **water vapor** at constant pressure
      (in :math:`J kg^-1 K^-1`).
    :rtype: float

    .. math::

      c_{pv} ~=~ 1858 + 3.820 \times 10^{-1} T + 4.220
      \times 10^{-4} T^2 - 1.996 \times 10^{-7} T^3

    "should be accurate for all near-surface atmospheric temperatures"
    [And2006]_.

    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    #
    th = _to_C(t, Kelvin=Kelvin)
    cp = 1858 + 3.820E-1*th + 4.220E-4 * (th**2) - 1.996E-7*(th**3)
    return cp

# ----------------------------------------------------------------------


def cwt(t, Kelvin=None):
    r'''
    Specific heat capacity of liquid water, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.

    :return: specific heat capacity of **liquid water** at constant pressure
      (in :math:`J kg^-1 K^-1`).
    :rtype: float

    .. math::

      c_{w} ~=~ 4217.4 - 3.720283 T^2 - 2.654387
      \times 10^{-3} T^3 + 2.093236 \times 10^{-5} T^4

    "should be accurate for all near-surface atmospheric temperatures"
    [And2006]_.

    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    #
    th = _to_C(t, Kelvin=Kelvin)
    cw = 4217.4 - 3.720283 * th**2 - 2.654387E-3 * th**3 + 2.093236E-5 * th**4
    return cw

# ----------------------------------------------------------------------


def cit(t, Kelvin=None):
    r'''
    Specific heat capacity of ice, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.

    :return: specific heat capacity of **ice** at constant pressure
      (in :math:`J kg^-1 K^-1`).
    :rtype: float

    .. math::

      c_{w} ~=~ -114.19 + 8.1288 T + 3.421 T~
      \exp \left[ -(T/125.1)^2 \right]

    "should be accurate for all near-surface atmospheric temperatures"
    [And2006]_.

    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    #
    th = _to_K(t, Kelvin=Kelvin)
    # here : th in K !
    ci = -114.19 + 8.1288 * th + 3.421 * th * np.exp(-(th/125.1)**2)
    return ci

# ----------------------------------------------------------------------


def Lvt(t, Kelvin=None, source="andreas"):
    r'''
    Latent heat of vaporization, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.
    :param source: (optional) select source of equation, see below. Defaults to
      `andreas`.

    :return: latent heat of vaporization (in :math:`J kg^-1`).
    :rtype: float

    Sources:

    andreas
      (**default**)
      [And2006]_ gives for Use in Marine Meteorology and states
      "the Lv values are within 0.3% of the Smithsonian
      values for temperatures from 0° to 60°C".

      :math:`L_{v} ~=~ `
      :math:`\left( 25.00 - 0.02274 ~ \vartheta \right) \times 10^{5}`

    henderson-sellers
      A different for of fit is given by [HeS1984]_:

      :math:`L_{v} ~=~ 1.91846E6 \left(T/\left(T-33.91\right)\right)^2`

    ecpack
      ecpack by Arain van Dijk and Arnold Moene uses:

      :math:`L_{v} ~=~ \left(2501 - 2.375 \vartheta\right) 10^3`

    dake
      a linear fit to the Smithonian tables is presented by [Dak1972]_:

      :math:`L_{v} ~=~ \left(2501 - 2.357*\vartheta\right) 10^3`

    hyland-wexler
      The American Society of Heating, Refrigerating and
      Air-Conditioning Engineers
      in their "ASHRAE Handbook : Fundamentals" presents values
      calculated using the formulation of [HyW1983]_ which evaluates to:

      .. math::
        :nowrap:

        \begin{eqnarray*}
        L_{v} &=& 3139.817121 \\
              &-& 2.390755077 * T \\
              &+& 4.3309357163E-4 * T^2 \\
              &-& 1.573317502E-6 * T^3 \\
              &+& 2.94377462234E-9 * T^4 \\
              &-& 1.7508262E-12 * T^5 \\
              &+& 6.059E-7 * (T - 403.128)^3 \\
        \end{eqnarray*}

    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    source = _check('source', source, 'str')
    # -----
    # Andreas (2006)
    # Physical Constants and Functions For Use in Marine Meteorology
    # Lv = ( 25.00 - 0.02274 * th ) * 1.0E5
    if source == "andreas":
        th = _to_C(t, Kelvin=Kelvin)
        Lv = (25.00 - 0.02274 * th) * 1.0E5
    # -----
    # Henderson‐Sellers (1984) DOI: 10.1002/qj.49711046626
    # T in K !
    # lv = 1.91846E6*(T/(T-33.91))**2
    elif source == "henderson-sellers":
        T = _to_K(t, Kelvin=Kelvin)
        Lv = 1.91846E6*(T/(T-33.91))**2
    # -----
    # Arnold Moene (EC-Pack). Source?
    # Lv = (2501 - 2.375*th)*1.0E3
    elif source == "ecpack":
        th = _to_C(t, Kelvin=Kelvin)
        Lv = (2501 - 2.375 * th) * 1.0E3
    # identical to Dake (1972) except typo
    # (there coefficient 0.5631 cal/K² -> 2.357 J/K²)
    # citation given by Drake is Simithonian tables (List 1951)
    # but there are only tabbed values "calculated after Goff_Gratch (1945)"
    elif source == "dake":
        th = _to_C(t, Kelvin=Kelvin)
        Lv = (2501 - 2.357 * th) * 1.0E3
    # -----
    # Goff_Gratch (1945) is not available
    # -----
    # (ASHRAE 2013) presents tabbed values calculated after
    # (Hyland and Wexler 2013), (ASHRAE 2013) are identical to
    # (List 1951) values, converted using 4.1839 J/cal
    elif source == "hyland-wexler":
        T = _to_K(t, Kelvin=Kelvin)
        Lv = (
            3139.817121 +
            -2.390755077 * T +
            4.3309357163E-4 * T**2 +
            -1.573317502E-6 * T**3 +
            2.94377462234E-9 * T**4 +
            -1.7508262E-12 * T**5 +
            6.059E-7 * (T - 403.128)**3
        ) * 1.0E3
    else:
        raise ValueError('evaporation heat source {} unknown'.format(source))

    return Lv

# ----------------------------------------------------------------------


def Lst(t, Kelvin=None, source="andreas"):
    r'''
    Latent heat of sublimation, as function of temperature

    :param t:  air temperature in :math:`K` or :math:`^\circ C`,
      depending on `Kelvin`.
    :param Kelvin: (optional) if ``False``, all temperatures are assumed to
      be Kelvin. If ``False``, all temperatures are assumed to be Celsius.
      If missing of ``None``, unit  temperatures are autodetected. Defaults to
      ``None``.
    :param source: (optional) select source of equation, see below. Defaults to
      `andreas`.

    :return: latent heat of sublimation (in :math:`J kg^-1`).
    :rtype: float

    Sources:

    andreas
      (**default**)
      [And2006]_ gives for Use in Marine Meteorology and states
      "the Ls values are within 0.2% of the Smithsonian
      values for temperatures from -50° to 0°C".
      :math:`L_{s} ~=~ \left( 28.34 - 0.00149 T \right) \times 10^{5}`

    hyland-wexler
      The American Society of Heating, Refrigerating and
      Air-Conditioning Engineers
      in their "ASHRAE Handbook : Fundamentals" presents values
      calculated using the formulation of [HyW1983]_ which evaluates to:

      .. math::
        :nowrap:

         \begin{eqnarray*}
         L_{v} &=&  2645.475 \\
           &+& 1.5292786 * T \\
           &-& 0.00254657837 * T^2 \\
           &-& 2.5511992E-6 * T^3 \\
           &+& 2.8726608E-9 * T^4 \\
           &-& 1.7508262E-12 * T^5
         \end{eqnarray*}


    '''
    t = _check('t', t, 'float', ge=0.)
    Kelvin = _check('Kelvin', Kelvin, 'bool', none=True)
    source = _check('source', source, 'str')
    # -----
    # Andreas (2006)
    # Physical Constants and Functions For Use in Marine Meteorology
    # Lv = ( 25.00 - 0.02274 * th ) * 1.0E5
    if source == "andreas":
        th = _to_C(t, Kelvin=Kelvin)
        Ls = (28.34 - 0.00149 * th) * 1.0E5
    # -----
    # (ASHRAE 2013) presents tabbed values calculated after
    # (Hyland and Wexler 2013), (ASHRAE 2013) are identical to
    # (List 1951) values, converted using 4.1839 J/cal
    elif source == "hyland-wexler":
        T = _to_K(t, Kelvin=Kelvin)
        Ls = (
            2645.475 +
            1.5292786 * T +
            -0.00254657837 * T**2 +
            -2.5511992E-6 * T**3 +
            2.8726608E-9 * T**4 +
            -1.7508262E-12 * T**5
        ) * 1.0E3
    else:
        raise ValueError('sublimation heat source {} unknown'.format(source))

    return Ls
